##############################################################################
#
# Copyright (c) 2006-2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""grokcore.catalog meta
"""
import martian
import zope.component
import grokcore.component
import grokcore.site
import grokcore.site.interfaces

from grokcore.catalog.components import IndexesClass
from martian.error import GrokError
from zope.catalog.catalog import Catalog
from zope.catalog.interfaces import ICatalog
from zope.exceptions.interfaces import DuplicationError
from zope.intid import IntIds
from zope.intid.interfaces import IIntIds
from zope.lifecycleevent.interfaces import IObjectAddedEvent


class IndexesGrokker(martian.InstanceGrokker):
    """Grokker for index bundles."""
    martian.component(IndexesClass)

    def grok(self, name, factory, module_info, config, **kw):
        site = grokcore.site.site.bind().get(factory)
        context = grokcore.component.context.bind().get(
            factory, module_info.getModule())
        install_on = grokcore.site.install_on.bind(
            default=IObjectAddedEvent).get(factory)
        catalog_name = grokcore.component.name.bind().get(factory)

        if site is None:
            raise GrokError(
                "No site specified for grokcore.catalog.Indexes "
                "subclass in module %r. "
                "Use grokcore.site.site() to specify."
                % module_info.getModule(), factory)

        indexes = getattr(factory, '__grok_indexes__', None)
        if indexes is None:
            return False

        subscriber = IndexesSetupSubscriber(
            catalog_name, indexes, context, module_info)

        subscribed = (site, install_on)
        config.action(
            discriminator=None,
            callable=grokcore.component.provideHandler,
            args=(subscriber, subscribed))
        return True


class IndexesSetupSubscriber(object):
    """Helper that sets up indexes when their Grok site is created.

    Each `grokcore.catalog.Indexes` class serves as an assertion that,
    whenever an instance of its `grokcore.site.site()` is created,
    the given list of indexes should be generated as well.

    But a long period of time could elapse between when the application
    starts (and its indexes are grokked), and the moment, maybe days or
    weeks later, when a new instance of that `grokcore.site.Site` is created.

    Hence this `IndexesSetupSubscriber`:
    it can be instantiated at grokking time with the index information,
    and then registered with the Component Architecture as an event that
    should be fired later, whenever the right kind of `grok.Site` is
    instantiated.  At that point its `__call__` method is kicked off and
    it makes sure the index catalogs get created properly.
    """
    def __init__(self, catalog_name, indexes, context, module_info):
        self.catalog_name = catalog_name
        self.indexes = indexes
        self.context = context
        self.module_info = module_info

    def __call__(self, site, event):
        # make sure we have an intids
        self._createIntIds(site)

        # get the catalog
        catalog = self._createCatalog(site)

        # now install indexes
        for name, index in self.indexes.items():
            try:
                index.setup(catalog, name, self.context, self.module_info)
            except DuplicationError:
                raise GrokError(
                    "grokcore.catalog.Indexes in module %r causes "
                    "creation of catalog index %r in catalog %r, "
                    "but an index with that name is already present." %
                    (self.module_info.getModule(), name, self.catalog_name),
                    None)

    def _createCatalog(self, site):
        """Create the catalog if needed and return it.

        If the catalog already exists, return that.

        """
        catalog = zope.component.queryUtility(
            ICatalog, name=self.catalog_name, context=site, default=None)
        if catalog is not None:
            return catalog
        catalog = Catalog()
        setupUtility = zope.component.getUtility(
            grokcore.site.interfaces.IUtilityInstaller)
        setupUtility(site, catalog, ICatalog, name=self.catalog_name)
        return catalog

    def _createIntIds(self, site):
        """Create intids if needed, and return it.
        """
        intids = zope.component.queryUtility(
            IIntIds, context=site, default=None)
        if intids is not None:
            return intids
        intids = IntIds()
        setupUtility = zope.component.getUtility(
            grokcore.site.interfaces.IUtilityInstaller)
        setupUtility(site, intids, IIntIds)
        return intids



"""
We now demonstrate the use of a ValueIndex with Grok::

Let's set up a site in which we manage a couple of objects::

  >>> herd = Herd()
  >>> getRootFolder()['herd'] = herd
  >>> from zope.site.hooks import setSite
  >>> setSite(herd)

Now we add some indexable objects to the site::

  >>> herd['alpha'] = Opossum('Alpha', datetime.datetime(2001, 6, 7))
  >>> herd['beta'] = Opossum('Beta', datetime.datetime(2001, 7, 8))
  >>> herd['gamma'] = Opossum('Gamma', datetime.datetime(2001, 7, 9))
  >>> herd['delta'] = Opossum('Delta', datetime.datetime(2001, 9, 24))

Let's query the datetime index::

  >>> from zope.catalog.interfaces import ICatalog
  >>> from zope.component import getUtility, queryUtility
  >>> catalog = getUtility(ICatalog)
  >>> def sortedResults(catalog, **kw):
  ...    result = list(catalog.searchResults(**kw))
  ...    result.sort(key=lambda x:x.name)
  ...    return [item.name for item in result]
  >>> sortedResults(catalog, birthday={
  ...     'any_of': [datetime.datetime(2001, 6, 7),
  ...                datetime.datetime(2001, 9, 24)]})
  ['Alpha', 'Delta']
  >>> sortedResults(catalog, birthday={
  ...     'between': [datetime.datetime(2001, 6, 7),
  ...                datetime.datetime(2001, 7, 9)]})
  ['Alpha', 'Beta', 'Gamma']
  >>> sortedResults(catalog, birthday={
  ...     'between': [datetime.datetime(2001, 6, 7),
  ...                datetime.datetime(2001, 7, 9),
  ...                True,  # exclude boundary.
  ...                True  # exclude boundary.
  ...                ]})
  ['Beta']

Note how the index uses a seconds-resolution and sub-second data is ignored::

  >>> herd['omega'] = Opossum(
  ...     'Omega', datetime.datetime(2002, 5, 3, 14, 53, 14, 1))
  >>> herd['psi'] = Opossum(
  ...     'Psi', datetime.datetime(2002, 5, 3, 14, 53, 14, 999999))
  >>> sortedResults(catalog, birthday={
  ...     'between': [datetime.datetime(2002, 5, 3, 14, 53, 14),
  ...                 datetime.datetime(2002, 5, 3, 14, 53, 14)]})
  ['Omega', 'Psi']

Nuke the catalog and intids in the end, so as not to confuse
other tests::

  >>> sm = herd.getSiteManager()
  >>> from zope.catalog.interfaces import ICatalog
  >>> sm.unregisterUtility(catalog, provided=ICatalog)
  True
  >>> from zope.intid.interfaces import IIntIds
  >>> from zope import component
  >>> intids = component.getUtility(IIntIds)
  >>> sm.unregisterUtility(intids, provided=IIntIds)
  True

Unfortunately ftests don't have good isolation from each other yet.
"""

import datetime
import grokcore.site
import grokcore.catalog
from grokcore.content import Container, Model
from zope.interface import implementer, Interface, Attribute


class Herd(Container, grokcore.site.Application):
    pass


class IOpossum(Interface):
    name = Attribute('Name')
    birthday = Attribute('Birthday')


class OpossumIndexes(grokcore.catalog.Indexes):
    grokcore.site.site(Herd)
    grokcore.catalog.context(IOpossum)

    birthday = grokcore.catalog.Datetime()


@implementer(IOpossum)
class Opossum(Model):

    def __init__(self, name, birthday):
        self.name = name
        self.birthday = birthday



"""
You can create an index on module level, but that should lead to a GrokError:

(Note how the test output needs to be on one line to please to
IGNORE_EXCEPTION_MODULE_IN_PYTHON2 normalizer)

  >>> func()
  Traceback (most recent call last):
    ...
  martian.error.GrokImportError: <class 'grokcore.catalog.index.Field'> can only be instantiated on class level.

  >>> from grokcore.catalog import testing
  >>> testing.grok(__name__)

"""
import grokcore.site
import grokcore.content
import grokcore.catalog
import zope.index.field

from grokcore.catalog import index, Indexes


def func():
    index.Field()


class Herd(grokcore.site.Application):
    pass


class Mammoth(grokcore.content.Model):
    pass


class FooIndex(zope.index.field.FieldIndex):
    # Not an IAttributeIndex index type.
    index_class = zope.index.field.index.FieldIndex


class FooIndexes(Indexes):
    grokcore.catalog.context(Mammoth)
    grokcore.catalog.name('foo_catalog')
    grokcore.site.site(Herd)

    foo = FooIndex()



"""
Grok allows you to set up catalog indexes in your application with a
special indexes declaration. Here we see how we can register indexes for
an interface instead of an application directly.

Let's set up a site in which we manage a couple of objects::

  >>> herd = Herd()
  >>> getRootFolder()['herd'] = herd
  >>> from zope.site.hooks import setSite
  >>> setSite(herd)

We are able to find the catalog::

  >>> from zope.catalog.interfaces import ICatalog
  >>> from zope.component import getUtility
  >>> catalog = getUtility(ICatalog)
  >>> catalog is not None
  True
  >>> catalog.get('name') is not None
  True

Nuke the catalog and intids for this site, so as not to confuse
other tests::

  >>> sm = herd.getSiteManager()
  >>> from zope.catalog.interfaces import ICatalog
  >>> sm.unregisterUtility(catalog, provided=ICatalog)
  True
  >>> from zope.intid.interfaces import IIntIds
  >>> from zope import component
  >>> intids = component.getUtility(IIntIds)
  >>> sm.unregisterUtility(intids, provided=IIntIds)
  True

Now let's create another application providing the same interface::

  >>> herd2 = Herd2()
  >>> getRootFolder()['herd2'] = herd2
  >>> setSite(herd2)
  >>> catalog = getUtility(ICatalog)
  >>> catalog is not None
  True
  >>> catalog.get('name') is not None
  True

Nuke the catalog and intids in the end, so as not to confuse
other tests::

  >>> sm = herd2.getSiteManager()
  >>> sm.unregisterUtility(catalog, provided=ICatalog)
  True
  >>> intids = component.getUtility(IIntIds)
  >>> sm.unregisterUtility(intids, provided=IIntIds)
  True
"""

import grokcore.site
import grokcore.catalog
from grokcore.content import Container
from zope.interface import Attribute, Interface, implementer


class IHerd(Interface):
    pass


@implementer(IHerd)
class Herd(Container, grokcore.site.Application):
    pass


@implementer(IHerd)
class Herd2(Container, grokcore.site.Application):
    pass


class IMammoth(Interface):
    name = Attribute("")
    age = Attribute("")

    def message():
        """Message the mammoth has for the world.
        """


class MammothIndexes(grokcore.catalog.Indexes):
    grokcore.site.site(IHerd)
    grokcore.catalog.context(IMammoth)

    name = grokcore.catalog.Field()
    age = grokcore.catalog.Field()
    message = grokcore.catalog.Text()



"""
Grok allows you to set up catalog indexes in your application with a
special indexes declaration. Here we show what happens if you try
to set up an index for an attribute that does not exist on the interface.

Let's set up a site in which we manage a couple of objects::

(Note how the test output needs to be on one line to please to
IGNORE_EXCEPTION_MODULE_IN_PYTHON2 normalizer)

  >>> herd = Herd()
  >>> getRootFolder()['herd'] = herd
  Traceback (most recent call last):
    ...
  martian.error.GrokError: grokcore.catalog.Indexes in <module 'grokcore.catalog.ftests.catalog.indexes_nonexistent' from ...> refers to an attribute or method 'foo' on interface <InterfaceClass grokcore.catalog.ftests.catalog.indexes_nonexistent.IMammoth>, but this does not exist.

Nuke the catalog and intids in the end, so as not to confuse
other tests::

  >>> from zope.site.hooks import setSite
  >>> setSite(herd)
  >>> from zope.catalog.interfaces import ICatalog
  >>> from zope.component import getUtility
  >>> catalog = getUtility(ICatalog)

  >>> sm = herd.getSiteManager()
  >>> from zope.catalog.interfaces import ICatalog
  >>> sm.unregisterUtility(catalog, provided=ICatalog)
  True
  >>> from zope.intid.interfaces import IIntIds
  >>> from zope import component
  >>> intids = component.getUtility(IIntIds)
  >>> sm.unregisterUtility(intids, provided=IIntIds)
  True
"""

import grokcore.site
import grokcore.catalog
from zope.interface import Interface, Attribute
from grokcore.content import Container


class Herd(Container, grokcore.site.Application):
    pass


class IMammoth(Interface):
    name = Attribute('name')
    age = Attribute('age')

    def message():
        """Message the mammoth has for the world.
        """


class MammothIndexes(grokcore.catalog.Indexes):
    grokcore.site.site(Herd)
    grokcore.catalog.context(IMammoth)
    foo = grokcore.catalog.Field()



import doctest
import unittest
from pkg_resources import resource_listdir

from zope.app.appsetup.testlayer import ZODBLayer
from zope.testing import renormalizing

import grokcore.catalog


FunctionalLayer = ZODBLayer(grokcore.catalog)


checker = renormalizing.RENormalizing()


def suiteFromPackage(name):
    files = resource_listdir(__name__, name)
    suite = unittest.TestSuite()
    for filename in files:
        if not filename.endswith('.py'):
            continue
        if filename == '__init__.py':
            continue

        dottedname = 'grokcore.catalog.ftests.%s.%s' % (name, filename[:-3])
        test = doctest.DocTestSuite(
            dottedname,
            checker=checker,
            extraglobs=dict(getRootFolder=FunctionalLayer.getRootFolder),
            optionflags=(doctest.ELLIPSIS +
                         doctest.NORMALIZE_WHITESPACE +
                         doctest.REPORT_NDIFF +
                         renormalizing.IGNORE_EXCEPTION_MODULE_IN_PYTHON2))
        test.layer = FunctionalLayer

        suite.addTest(test)
    return suite


def test_suite():
    suite = unittest.TestSuite()
    for name in ["catalog"]:
        suite.addTest(suiteFromPackage(name))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
